{-# LANGUAGE StandaloneDeriving, FlexibleInstances #-}

module Main ( main
            ) where

import Algos.Graphs.Forest(Forest(..), Tree(..), Subtree(..), preOrder)
import Algos.Graphs.Graph(Graph(..), Vertex(..), Edge(..))
import Algos.Graphs.GraphIO(Orientation(..), Display(..), Parse(..), parseGraph, NewickType(..))
import Control.Applicative ((<$>))
import Data.Array (Array, array, (!))
import Data.Monoid (Monoid(..))
import Data.Set (Set, fromList, member)
import System.Environment (getArgs, getProgName)
import System.Exit (exitFailure)

main :: IO ()
main =  do (gfile, ffile) <- readArgs
           (o, g) <- parseGraph  <$> readFile gfile
           f      <- parseNewick <$> readFile ffile
           putStrLn (formatResult $ verifyBFSForest o g f)

readArgs :: IO (String, String)
readArgs =  do args <- getArgs
               case args of
                 [gfile, ffile] -> return (gfile, ffile)
                 _              -> usage >> exitFailure

usage :: IO ()
usage =  do prog <- getProgName
            putStrLn $ "USAGE: " ++ prog ++ " <graph file> <forest file>"

formatResult                           :: Result -> String
formatResult OK                        =  "OK"
formatResult (MissingVertex v)         =  "Vertex " ++ show v ++ " is missing from forest"
formatResult (NotAnEdge e)             =  "Edge in forest is not in the graph: " ++ show e
formatResult (LongForwardEdge e tl hl) =  "Found forward edge from level " ++ show tl ++ 
                                          " to level " ++ show hl ++ ": " ++ show e
formatResult (LateParent v pv u)       =  "Node " ++ show v ++ " has parent " ++ show pv ++
                                          " but should have been a child of " ++ show u

data Result = OK
            | MissingVertex   Int                -- vertex
            | NotAnEdge       (Int, Int)         -- edge
            | LongForwardEdge (Int, Int) Int Int -- edge, tail level, head level
            | LateParent      Int Int Int        -- node, parent, earlier node

instance Monoid Result where
  mempty       = OK
  mappend OK x = x
  mappend x  _ = x

verifyBFSForest       :: Orientation -> Graph Int Int Int -> Forest Int Int -> Result
verifyBFSForest o g f =  mconcat [missingVertex, nonEdge, longForwardEdge, lateParent]

  where fvs           = forestVertices f
        missingVertex = mconcat $ map (checkIfPresent fvs) (graphVertices g)
          
        ges     = graphEdgeSet g
        nonEdge = mconcat $ map (checkIfGraphEdge o ges) (forestEdges f)
        
        n               = length (graphVertices g)
        levels          = vertexLevels n f
        comps           = components n f
        longForwardEdge = mconcat $ map (checkForLongEdge o levels comps) (graphEdges g)
        
        prnts           = parents n f
        preorderNumbers = array (1,n) (zip (preOrder f) [1..])
        lateParent      = mconcat $ 
                          map (checkForLateParent o prnts levels preorderNumbers)
                              (graphEdges g)

forestVertices :: Forest Int Int -> Set Int
forestVertices =  fromList . preOrder

forestEdges :: Forest Int Int -> [Edge Int Int]
forestEdges =  concatMap treeEdges . trees

treeEdges   :: Tree Int Int -> [Edge Int Int]
treeEdges t =  go t []
  where go (Tree _ []) es = es
        go (Tree v ws) es = map (\(Subtree e (Tree w _)) -> Edge v w e) ws ++
                            foldr go es (map (\(Subtree _ t) -> t) ws)

vertexLevels               :: Int -> Forest Int Int -> Array Int Int
vertexLevels n (Forest ts) =  array (1, n) (concatMap vertexLevels' ts)

vertexLevels'   :: Tree Int Int -> [(Int, Int)]
vertexLevels' t =  go 0 t []
  where go d (Tree v ws) ds = (v, d) : foldr (go (d+1)) ds (map (\(Subtree _ t) -> t) ws)

components               :: Int -> Forest Int Int -> Array Int Int
components n (Forest ts) =  array (1, n) $ concat $ zipWith (\vs i -> zip vs (repeat i)) (map preOrder ts) [1..]

parents     :: Int -> Forest Int Int -> Array Int Int
parents n f =  array (1, n) ps
  where ps = map (\(Edge pv v _) -> (v, pv)) (forestEdges f)

graphEdgeSet :: Graph Int Int Int -> Set (Edge Int Int)
graphEdgeSet =  fromList . graphEdges

checkIfPresent                 :: Set Int -> Vertex Int Int -> Result
checkIfPresent vs (Vertex v _) |  v `member` vs = OK
                               |  otherwise     = MissingVertex v

checkIfGraphEdge                     :: Orientation -> Set (Edge Int Int) -> Edge Int Int -> Result
checkIfGraphEdge o es e@(Edge u v l) |  e `member` es                               = OK
                                     |  o == Undirected && (Edge v u l) `member` es = OK
                                     |  otherwise                                   = NotAnEdge (u, v)

checkForLongEdge                      :: Orientation -> Array Int Int -> Array Int Int -> Edge Int Int -> Result
checkForLongEdge o ls cs (Edge u v _) |  abs (lu - lv) <= 1       = OK
                                      |  o == Directed && lv < lu = OK
                                      |  o == Directed && cv < cu = OK
                                      |  otherwise                = LongForwardEdge (u, v) lu lv
  where lu = ls ! u
        lv = ls ! v
        cu = cs ! u
        cv = cs ! v

checkForLateParent :: Orientation -> Array Int Int -> Array Int Int -> 
                      Array Int Int -> Edge Int Int -> Result
checkForLateParent o prnts levels preNums (Edge u v _) 
  |  o == Directed               && (lu > lpv || numu >= numpv) = OK
  |  (lu > lpv || numu >= numpv) && (lv > lpu || numv >= numpu) = OK
  |  otherwise                                                  = LateParent v pv u
  where pu | lu > 0    = prnts ! u
           | otherwise = u
        pv | lv > 0    = prnts ! v
           | otherwise = v
        lu  = levels ! u
        lv  = levels ! v
        lpu = levels ! pu
        lpv = levels ! pv
        numu  = preNums ! u
        numv  = preNums ! v
        numpu = preNums ! pu
        numpv = preNums ! pv

instance Display Int where
  display = show

instance Parse Int where
  parse = read

deriving instance Eq (Vertex Int Int)
deriving instance Eq (Edge Int Int)

deriving instance Ord (Vertex Int Int)
deriving instance Ord (Edge Int Int)
